package xyz.lychee.lagfixer.libs.adventure.text;

import java.util.regex.Matcher;
import java.util.stream.Stream;
import xyz.lychee.lagfixer.libs.annotations.Contract;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.ApiStatus.ScheduledForRemoval;
import xyz.lychee.lagfixer.libs.examination.Examinable;
import xyz.lychee.lagfixer.libs.examination.ExaminableProperty;

public interface BlockNBTComponent extends NBTComponent<BlockNBTComponent, Builder>, ScopedComponent<BlockNBTComponent> {
   @NotNull
   @NotNull Pos pos();

   @Contract(
      pure = true
   )
   @NotNull BlockNBTComponent pos(@NotNull final @NotNull Pos pos);

   @Contract(
      pure = true
   )
   default @NotNull BlockNBTComponent localPos(final double left, final double up, final double forwards) {
      return this.pos(BlockNBTComponent.LocalPos.localPos(left, up, forwards));
   }

   @Contract(
      pure = true
   )
   default @NotNull BlockNBTComponent worldPos(final WorldPos.@NotNull Coordinate x, final WorldPos.@NotNull Coordinate y, final WorldPos.@NotNull Coordinate z) {
      return this.pos(BlockNBTComponent.WorldPos.worldPos(x, y, z));
   }

   @Contract(
      pure = true
   )
   default @NotNull BlockNBTComponent absoluteWorldPos(final int x, final int y, final int z) {
      return this.worldPos(BlockNBTComponent.WorldPos.Coordinate.absolute(x), BlockNBTComponent.WorldPos.Coordinate.absolute(y), BlockNBTComponent.WorldPos.Coordinate.absolute(z));
   }

   @Contract(
      pure = true
   )
   default @NotNull BlockNBTComponent relativeWorldPos(final int x, final int y, final int z) {
      return this.worldPos(BlockNBTComponent.WorldPos.Coordinate.relative(x), BlockNBTComponent.WorldPos.Coordinate.relative(y), BlockNBTComponent.WorldPos.Coordinate.relative(z));
   }

   default @NotNull Stream<? extends ExaminableProperty> examinableProperties() {
      return Stream.concat(Stream.of(ExaminableProperty.of("pos", this.pos())), NBTComponent.super.examinableProperties());
   }

   public interface WorldPos extends Pos {
      @NotNull
      static @NotNull WorldPos worldPos(@NotNull final @NotNull Coordinate x, @NotNull final @NotNull Coordinate y, @NotNull final @NotNull Coordinate z) {
         return new BlockNBTComponentImpl.WorldPosImpl(x, y, z);
      }

      /** @deprecated */
      @Deprecated
      @ScheduledForRemoval(
         inVersion = "5.0.0"
      )
      @NotNull
      static @NotNull WorldPos of(@NotNull final @NotNull Coordinate x, @NotNull final @NotNull Coordinate y, @NotNull final @NotNull Coordinate z) {
         return new BlockNBTComponentImpl.WorldPosImpl(x, y, z);
      }

      @NotNull
      @NotNull Coordinate x();

      @NotNull
      @NotNull Coordinate y();

      @NotNull
      @NotNull Coordinate z();

      public interface Coordinate extends Examinable {
         @NotNull
         static @NotNull Coordinate absolute(final int value) {
            return coordinate(value, BlockNBTComponent.WorldPos.Coordinate.Type.ABSOLUTE);
         }

         @NotNull
         static @NotNull Coordinate relative(final int value) {
            return coordinate(value, BlockNBTComponent.WorldPos.Coordinate.Type.RELATIVE);
         }

         @NotNull
         static @NotNull Coordinate coordinate(final int value, @NotNull final @NotNull Type type) {
            return new BlockNBTComponentImpl.WorldPosImpl.CoordinateImpl(value, type);
         }

         /** @deprecated */
         @Deprecated
         @ScheduledForRemoval(
            inVersion = "5.0.0"
         )
         @NotNull
         static @NotNull Coordinate of(final int value, @NotNull final @NotNull Type type) {
            return new BlockNBTComponentImpl.WorldPosImpl.CoordinateImpl(value, type);
         }

         int value();

         @NotNull
         @NotNull Type type();

         public static enum Type {
            ABSOLUTE,
            RELATIVE;
         }
      }
   }

   public interface LocalPos extends Pos {
      @NotNull
      static @NotNull LocalPos localPos(final double left, final double up, final double forwards) {
         return new BlockNBTComponentImpl.LocalPosImpl(left, up, forwards);
      }

      /** @deprecated */
      @Deprecated
      @ScheduledForRemoval(
         inVersion = "5.0.0"
      )
      @NotNull
      static @NotNull LocalPos of(final double left, final double up, final double forwards) {
         return new BlockNBTComponentImpl.LocalPosImpl(left, up, forwards);
      }

      double left();

      double up();

      double forwards();
   }

   public interface Pos extends Examinable {
      @NotNull
      static @NotNull Pos fromString(final @NotNull String input) throws IllegalArgumentException {
         Matcher localMatch = BlockNBTComponentImpl.Tokens.LOCAL_PATTERN.matcher(input);
         if (localMatch.matches()) {
            return BlockNBTComponent.LocalPos.localPos(Double.parseDouble(localMatch.group(1)), Double.parseDouble(localMatch.group(3)), Double.parseDouble(localMatch.group(5)));
         } else {
            Matcher worldMatch = BlockNBTComponentImpl.Tokens.WORLD_PATTERN.matcher(input);
            if (worldMatch.matches()) {
               return BlockNBTComponent.WorldPos.worldPos(BlockNBTComponentImpl.Tokens.deserializeCoordinate(worldMatch.group(1), worldMatch.group(2)), BlockNBTComponentImpl.Tokens.deserializeCoordinate(worldMatch.group(3), worldMatch.group(4)), BlockNBTComponentImpl.Tokens.deserializeCoordinate(worldMatch.group(5), worldMatch.group(6)));
            } else {
               throw new IllegalArgumentException("Cannot convert position specification '" + input + "' into a position");
            }
         }
      }

      @NotNull String asString();
   }

   public interface Builder extends NBTComponentBuilder<BlockNBTComponent, Builder> {
      @Contract("_ -> this")
      @NotNull
      @NotNull Builder pos(@NotNull final @NotNull Pos pos);

      @Contract("_, _, _ -> this")
      @NotNull
      default @NotNull Builder localPos(final double left, final double up, final double forwards) {
         return this.pos(BlockNBTComponent.LocalPos.localPos(left, up, forwards));
      }

      @Contract("_, _, _ -> this")
      @NotNull
      default @NotNull Builder worldPos(final WorldPos.@NotNull Coordinate x, final WorldPos.@NotNull Coordinate y, final WorldPos.@NotNull Coordinate z) {
         return this.pos(BlockNBTComponent.WorldPos.worldPos(x, y, z));
      }

      @Contract("_, _, _ -> this")
      @NotNull
      default @NotNull Builder absoluteWorldPos(final int x, final int y, final int z) {
         return this.worldPos(BlockNBTComponent.WorldPos.Coordinate.absolute(x), BlockNBTComponent.WorldPos.Coordinate.absolute(y), BlockNBTComponent.WorldPos.Coordinate.absolute(z));
      }

      @Contract("_, _, _ -> this")
      @NotNull
      default @NotNull Builder relativeWorldPos(final int x, final int y, final int z) {
         return this.worldPos(BlockNBTComponent.WorldPos.Coordinate.relative(x), BlockNBTComponent.WorldPos.Coordinate.relative(y), BlockNBTComponent.WorldPos.Coordinate.relative(z));
      }
   }
}
